"""Plotting module for visualizing how well the GP fits the data."""

from collections.abc import Callable
from typing import TypeAlias

import matplotlib.pyplot as plt
import numpy as np
import plotly
import plotly.graph_objects
import plotly.subplots
import torch
from ax.core import ObservationFeatures, RangeParameter, SearchSpace
from ax.modelbridge.torch import TorchModelBridge
from botorch.models import SingleTaskGP
from botorch.posteriors.gpytorch import GPyTorchPosterior
from matplotlib.axes import Axes
from plotly.graph_objs import Figure, Scatter3d, Surface
from scipy.stats import norm

Numpy2dArray: TypeAlias = np.ndarray[tuple[int, int], np.dtype[np.float64]]
Numpy1dArray: TypeAlias = np.ndarray[tuple[int,], np.dtype[np.float64]]


def plot_surface_over_2d_search_space(
    search_space: SearchSpace,
    funcs: list[Callable[[Numpy2dArray], Numpy1dArray]],
    colors: list[str] | None = None,
    num_points: int = 101,
) -> Figure:
    """Creates a figure with the functions in funcs ploted over the search_space.

    Note:
        Currently only support search spaces with 2 parameters.

    Args:
        search_space: The search space over which the functions are to be evaluated and plotted.
        funcs: A list of callables that take in a numpy array with shape (num_values, num_parameters=2 )
            and return a numpy array with (num_values) elements.
        colors: A list of colors to use for each function. If None, will use default Plotly colors.
        num_points: The number of points in each dimension to evaluate the functions at.
    """
    # Extract the parameter names and ranges from the search space
    assert len(search_space.parameters) == 2, "Only 2D search spaces are supported for now."  # noqa: PLR2004

    (x1_name, x1_param), (x2_name, x2_param) = list(search_space.parameters.items())

    if not (isinstance(x1_param, RangeParameter) and isinstance(x2_param, RangeParameter)):
        msg = f"""Expect search_space.parameters to all be of type RangeParameter.
         Instead got {type(x1_param) = }, and {type(x2_param) = }."""
        raise NotImplementedError(msg)

    # Generate parameter ranges using NumPy
    x1_values = np.linspace(x1_param.lower, x1_param.upper, num_points)
    x2_values = np.linspace(x2_param.lower, x2_param.upper, num_points)

    # Create a meshgrid for the parameter values

    x1_grid, x2_grid = np.meshgrid(x1_values, x2_values)
    inputs = np.column_stack((x1_grid.flatten(), x2_grid.flatten()))

    func_results = [func(inputs) for func in funcs]

    # Create a 3D surface plot using Plotly
    surfaces = []
    _none_list: list[None] = [None] * len(funcs)
    colour_list: list[str] | list[None] = colors or _none_list
    for func_result, color in zip(func_results, colour_list, strict=True):
        surface = plotly.graph_objects.Surface(
            x=x1_values,
            y=x2_values,
            z=func_result.reshape(x1_grid.shape),
            opacity=0.7,
            colorscale=color,
        )
        surfaces.append(surface)

    return plotly.graph_objects.Figure(data=surfaces)


def scatter_plot_training(
    model_bridge: TorchModelBridge,
    metric_name: str,
    axis: tuple[int, int] = (0, 1),
    figure: Figure | None = None,
    *,
    error_bars: bool = True,
    error_bar_confidence_interval: float = 0.95,
) -> Figure:
    """Make a scattter plot of a metric for the training data of the model.

    Args:
        model_bridge: The model the training data scatter plot is to be made for.
        metric_name: The name of the metric to plot. Must match the name of a metric in the model.
        axis: The axis of the input space to plot the scatter plot in
        figure: The figure to add the scatter plot to. If None, a new figure is created.
        error_bars: Whether to add error bars to the plot.
        error_bar_confidence_interval: The confidence interval the error bars in the scatter plot represents.

    Returns:
        A Scatter3d plot of the training data for the given metric.
    """
    xs = np.array([list(obs.features.parameters.values()) for obs in model_bridge.get_training_data()])
    ys = []
    y_noise = []
    for obs in model_bridge.get_training_data():
        metric_index = obs.data.metric_names.index(metric_name)
        ys.append(obs.data.means[metric_index])
        y_noise.append(np.sqrt(obs.data.covariance[metric_index][metric_index]))

    figure = figure or plotly.graph_objects.Figure()
    scatter = Scatter3d(
        x=xs[:, axis[0]],
        y=xs[:, axis[1]],
        z=ys,
        mode="markers+text",
        marker={"size": 5, "color": "red"},
    )

    _ = figure.add_trace(scatter)
    if error_bars:
        if error_bar_confidence_interval >= 1.0:
            error_bar_confidence_interval = 1.0 - 1e-6
        elif error_bar_confidence_interval <= 0.0:
            error_bar_confidence_interval = 0.0

        standard_interval = norm.interval(error_bar_confidence_interval)
        # Add error bars
        for i in range(xs.shape[0]):
            _ = figure.add_trace(
                plotly.graph_objects.Scatter3d(
                    x=[
                        xs[i, axis[0]],
                        xs[i, axis[0]],
                    ],
                    y=[
                        xs[i, axis[1]],
                        xs[i, axis[1]],
                    ],
                    z=[
                        ys[i] + y_noise[i] * standard_interval[0],
                        ys[i] + y_noise[i] * standard_interval[1],
                    ],
                    mode="lines",
                    line={"color": "red"},
                )
            )
    return figure


def plot_gp_fits_2d_surface(  # noqa: C901
    model_bridge: TorchModelBridge,
    search_space: SearchSpace,
    metrics: dict[
        str,
        Callable[[Numpy2dArray], Numpy1dArray],
    ]
    | None = None,
    num_points: int = 101,
) -> Figure:
    """Plot the GP fit for the given metrics over the 2D search space.

    Args:
        model_bridge: The model bridge used to make predictions.
        search_space: The search space over which the functions are to be evaluated and plotted.
        metrics: A dictionary of metrics to plot. The keys are the names of the metrics in the model bridge model
            and the values are callables that return the metric value for a given input.
        num_points: The number of points in each dimension to evaluate the functions at.
    """
    # Extract the parameter names and ranges from the search space

    figs = []
    observations_names = search_space.parameters.keys()

    # Get metric names from the model
    metrics_names = list(model_bridge.metric_names)

    def pred_helper_closure(
        metric_name: str,
    ) -> Callable[[Numpy2dArray], Numpy1dArray]:
        def pred_helper(
            x: Numpy2dArray,
        ) -> Numpy1dArray:
            if len(x.shape) == 1:
                x = x.reshape(1, -1)
            pred_val_list = [
                model_bridge.predict(
                    [
                        ObservationFeatures(
                            parameters={
                                obs_name: x[i, obs_index] for obs_index, obs_name in enumerate(observations_names)
                            }
                        )
                    ]
                )[0][metric_name]
                for i in range(x.shape[0])
            ]
            return np.array(pred_val_list)

        return pred_helper

    for metric_name in metrics_names:
        fig = None
        funcs = [pred_helper_closure(metric_name)]
        colors = ["Reds"]

        # If we have a ground truth function for the metric add it to the plot
        if metrics is not None and metric_name in metrics:
            funcs.insert(0, metrics[metric_name])
            colors.insert(0, "Viridis")

        fig = plot_surface_over_2d_search_space(
            search_space=search_space,
            funcs=funcs,
            colors=colors,
            num_points=num_points,
        )

        fig = scatter_plot_training(
            model_bridge=model_bridge,
            metric_name=metric_name,
            figure=fig,
        )
        _ = fig.update_scenes(
            {
                "xaxis": {"title": "x1"},
                "yaxis": {"title": "x2"},
                "zaxis": {"title": "response"},
            }
        )
        figs.append(fig)

    ### Plot them side by side
    fig_subplots = plotly.subplots.make_subplots(
        rows=1,
        cols=len(metrics_names),
        specs=[[{"type": "surface"}, {"type": "surface"}]],
        subplot_titles=metrics_names,
    )

    for idx, fig in enumerate(figs):
        for data in fig.data:
            _ = fig_subplots.add_trace(data, row=1, col=idx + 1)

    _ = fig_subplots.update_scenes({"xaxis": {"title": "x1"}, "yaxis": {"title": "x2"}, "zaxis": {"title": "response"}})

    # Remove the colorbar and legend from the subplots
    for trace in fig_subplots.data:
        if isinstance(trace, Surface):
            _ = trace.update(showscale=False)
        elif isinstance(trace, Scatter3d):
            _ = trace.update(showlegend=False)
    return fig_subplots


def plot_1d_model(model: SingleTaskGP, X: torch.Tensor | None = None, ax: None | Axes = None) -> Axes:  # noqa: N803
    """Plots a model with 1d in put, and any number of outputs..

    Args:
        model: Only SingleTaskGp is supported an training data is extracted from the model.
        X: (n,1): Linspace of [0,1] is used by default. Only 1d is currently supported.
        ax: will plot to this axis if provied
    """
    X = X or torch.linspace(0, 1, 100).reshape(-1, 1)  # noqa: N806

    if ax is None:
        _, ax = plt.subplots()

    # Quick and dirty way to ensure consistent colours
    colours = ["b", "c", "m", "y", "k", "r"]

    train_x_all = model.train_inputs[0].detach()  # type: ignore  # noqa: PGH003
    train_y_all = model.train_targets.detach()
    train_var_all = model.likelihood.noise.detach()

    # t = 1 then:
    # shape model.train_targets: (n)
    # shape model.train_inputs: ((n,d),)
    if len(model.train_targets.shape) == 1:
        n_targets = 1
        # need to add a t dimension to all of these so they can be treated the same way as the multicase
        train_x_all = train_x_all.unsqueeze(0)
        train_y_all = train_y_all.unsqueeze(0)
        train_var_all = train_var_all.unsqueeze(0)

    # t > 1 then:
    # shape model.train_targets: (t,n)
    # shape model.train_inputs: ((t,n,d),)
    else:
        n_targets = model.train_targets.shape[-2]

    for target_idx in range(n_targets):
        c = colours[target_idx % len(colours)]

        train_x = train_x_all[target_idx]
        train_y = train_y_all[target_idx]
        train_var = train_var_all[target_idx]

        with torch.no_grad():
            posterior: GPyTorchPosterior = model.posterior(X)  # type: ignore  # noqa: PGH003

        mean = posterior.mean[:, target_idx]
        var = posterior.variance[:, target_idx]
        _ = ax.fill_between(X.flatten(), mean - 1.95 * var**0.5, mean + 1.95 * var**0.5, alpha=0.3, color=c)
        _ = ax.plot(X, mean, color=c, label=f"gp target {target_idx}")
        _ = ax.scatter(train_x.flatten(), train_y, color=c)
        _ = ax.errorbar(train_x.flatten(), train_y, 1.95 * train_var**0.5, fmt="o", color=c)

    _ = ax.set_title("Gp prediction")
    return ax



"""Helpers for unpacking nestes objects for logging purposes.

When running experiments it is useful to record the objects that produced those results for reproducibility. This module
provides helper for unpacking nested objects for logging. It default implementations, and supports cusomistion so the
right level of granuality can be achieved.
"""

# pyright: reportUnnecessaryTypeIgnoreComment=false

# %%
from collections.abc import Callable
from typing import Any, TypeAlias, TypeVar, Union

import torch

T = TypeVar("T")
# Nested dict where keys are alsways strings
NestedDict: TypeAlias = dict[str, Union[Any, "NestedDict"]]
NestedStrDict: TypeAlias = dict[str, Union[str, "NestedStrDict"]]


# %%
def unpack_object(
    obj: object,
    custom_unpacking_config: dict[type, Callable[[Any], NestedDict]] | None = None,
    depth: int = 1,
) -> NestedDict:
    """Recursively extracts attributes from objects.

    This can be useful for logging the state of an object. It will unpack the public attributes of an object up to
    'depth'. Specific unpacking function can also be provided for attribute object in `custom_unpacking_config`. If
    these objects are encounter the unpacking function will be used instead.

    Args:
        obj: the object to unpack.
        custom_unpacking_config: Overrides default unpacking behaviour for object that subclass the keys.

            - Keys: Types
            - Values: Functions that take instance of that type, and produces a custom unpacking.
            - This unpacking should be of the following format:

        depth:
            - How many levels of objects to unpack.

    Return:
        A nested dictionary.

        - Without custom unpacking:
            - keys: are the public attribute names
            - values: are the attribute value, or nested dictionary of the object being unpacked.

                {"__class__": RootObjectClass, "attribute_1": {"__class__": Foo, "a": "blah", "b": None}, "attribute_2": "x"}

        - With custom unpacking:
            - keys: Determined by the custom unpacking function if used otherwise as above.
            - Values: Determined by the custom unpacking function if used otherwise as above.

                {"__class__": RootObjectClass, "attribute_1": {"custom_unpacking_key1": "name of class if FOO", "custom_unpacking_key2": [1, 2, 3]}, "attribute_2": "x"}

    Todo:
        depth: How many levels of objects to unpack.
    """  # noqa: E501
    custom_unpacking_config = custom_unpacking_config or {}

    if item_unpacker := get_closest_class_value(obj, custom_unpacking_config):
        return item_unpacker(obj)
    # Base case: out of depth - not further processing
    if depth == 0:
        return {"__class__": type(obj)}
    # Basecase: if unpackable object do nothing
    if not hasattr(obj, "__dict__"):
        return {"__class__": type(obj)}

    # Recussion:
    # - find child items
    # - work out which ones can be processed
    attributes = public_vars(obj)

    # Process if: tthe child can be unpacked.
    # sourcery skip: dict-comprehension
    updated_attribute: dict[str, Any] = {}
    for attr_name, attr_value in attributes.items():
        # Check if can be expanded or there is a specific processing function.
        if hasattr(attr_value, "__dict__") or get_closest_class_value(attr_value, custom_unpacking_config):  # type: ignore[arg-type]
            updated_attribute[attr_name] = unpack_object(attr_value, custom_unpacking_config, depth=depth - 1)

    return {"__class__": type(obj), **attributes, **updated_attribute}


def get_closest_class_value(obj: object, dic: dict[type, T]) -> T | None:
    """Searches a dictionary for the closest class, and return the value stored.

    Searches the class heirachy from bottom to top for a matching key in the dictionary. Returns the values stored with
    by that key.

    Args:
        obj: Object to find the class for
        dic: takes object of that type,

    Return:
        The value in stored in the dictionary for the closest class, or None if no match is found.
    """
    return next((dic[cls] for cls in obj.__class__.mro() if cls in dic), None)


def public_vars(obj: object) -> dict[str, Any]:
    """Like `vars()` but just returns public items."""
    return {k: v for k, v in vars(obj).items() if (k[0] != "_")}


def nested_content_to_str(d: NestedDict) -> NestedStrDict:
    """Helper function to turn all values in nested dictionaries to `str`."""

    def convert(value: Any) -> NestedDict | str:  # noqa: ANN401
        # If the value is a dictionary, recursively apply the function
        if isinstance(value, dict):
            return {str(k): convert(v) for k, v in value.items()}
        return str(value)

    return convert(d)  # type: ignore  # noqa: PGH003


def unpack_object_str_content(
    obj: object, custom_unpacking_config: None | dict[type, Callable[[Any], NestedDict]] = None, depth: int = 1
) -> NestedStrDict:
    """Helper that converts all `unpack_object` to string.

    Args:
        obj: the object to unpack.
        custom_unpacking_config: Overrides default unpacking behaviour for object that subclass the keys. If `None`
            `default_config()` is used.

            - Keys: Types
            - Values: Functions that take instance of that type, and produces a custom unpacking. This unpacking should
              be of the following format:

        depth:
            - How many levels of objects to unpack.
    """
    if custom_unpacking_config is None:
        custom_unpacking_config = default_config()
    return nested_content_to_str(unpack_object(obj, custom_unpacking_config, depth))


def default_config() -> dict[type, Callable[[Any], NestedDict]]:
    """Default/handy processing configurations."""
    import numpy as np
    from botorch.models import SingleTaskGP

    def process_single_task_gp(x: SingleTaskGP) -> dict[str, Any]:
        return {
            "__class__": type(x),
            "train_inputs": x.train_inputs[0].shape,  # pyright: ignore[reportOptionalSubscript]
            "train_targets": x.train_targets.shape,
            "likelihood.noise": x.likelihood.noise.shape,
        }

    # Helper for producing output
    def class_and_shape(x):  # type: ignore  # noqa: ANN001, ANN202, PGH003
        return {"__class__": type(x), "shape": x.shape}

    return {torch.Tensor: class_and_shape, np.ndarray: class_and_shape, SingleTaskGP: process_single_task_gp}



"""Sequential sampling."""

from collections.abc import Iterator, Sized

from torch.utils.data import Sampler


class SizableSequentialSampler(Sampler[int]):
    """Samples elements sequentially, always in the same order.

    Follows the pattern in RandomSampler to allow for sampling a specific number of samplers.
    This can be smaller or larger than the amount in the dataset.
    """

    data_source: Sized  # this follow the patter in SequentialSampler

    def __init__(self, data_source: Sized, num_samples: None | int = None) -> None:
        """Create the sampler.

        Args:
            data_source (Dataset): dataset to sample from
            num_samples (int): number of samples to draw, default=`len(dataset)`.
        """
        self.data_source = data_source
        self._num_samples = num_samples

    @property
    def num_samples(self) -> int:  # noqa: D102
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self) -> Iterator[int]:  # noqa: D105
        n = len(self.data_source)

        for _ in range(self.num_samples // n):
            yield from iter(range(n))
        yield from iter(range(self.num_samples % n))

    def __len__(self) -> int:  # noqa: D105
        return self.num_samples



"""Additional Runner implementatiions."""

from collections.abc import Iterable
from typing import Any

from ax import Runner, Trial
from ax.core.base_trial import BaseTrial, TrialStatus

from axtreme.evaluation import EvaluationFunction


class LocalMetadataRunner(Runner):
    """Except for the addition of evauation_function this is coppied exactly from ax.runners.SyntheticRunner."""

    def __init__(self, evaluation_function: EvaluationFunction) -> None:  # noqa: D107
        self.evaluation_function = evaluation_function

    def run(self, trial: BaseTrial) -> dict[str, Any]:  # noqa: D102
        # The other available type is BatchTrail which is for very specific usecase not relevant to this project. see https://ax.dev/docs/core.html#trial-vs-batched-trial
        if isinstance(trial, Trial):
            deployed_name = (
                f"{trial.experiment.name}_{trial.index!s}" if trial.experiment.has_name else str(trial.index)
            )
            assert trial.arm is not None  # primarily to deal with typing
            simulation_result = self.evaluation_function(trial.arm.parameters)

            return {"name": deployed_name, "simulation_result": simulation_result}

        msg = f"Currently only trial of type Trial is supported. Received {type(trial)}."
        raise NotImplementedError(msg)

    def poll_trial_status(self, trials: Iterable[BaseTrial]) -> dict[TrialStatus, set[int]]:  # noqa: D102
        # only completed if simulation_result is available
        completed_trials = [t for t in trials if t.run_metadata.get("simulation_result") is not None]
        # TODO(magkri): not sure how to handle all TrialStatus values here, maybe we shouldn't override this method
        return {
            TrialStatus.COMPLETED: {t.index for t in completed_trials},
            TrialStatus.RUNNING: {t.index for t in trials if t not in completed_trials},
        }

    @property
    def run_metadata_report_keys(self) -> list[str]:  # noqa: D102
        return ["name", "simulation_result"]



"""The module is used to determine the ax transformations used and create the equivalent botorch transformation.

This allows the user to work directly with the Botorch model, while using inputs and outputs in the original space
(e.g the problem space)

Todo:
    It would be nice to be able to return an identity transform so we don't have to deal with Nones (sw 2024-11-21).
"""

from collections.abc import Callable
from functools import partial
from typing import Any, TypeAlias

import torch
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.cast import Cast
from ax.modelbridge.transforms.choice_encode import OrderedChoiceEncode
from ax.modelbridge.transforms.derelativize import Derelativize
from ax.modelbridge.transforms.int_to_float import IntToFloat
from ax.modelbridge.transforms.ivw import IVW
from ax.modelbridge.transforms.log import Log
from ax.modelbridge.transforms.logit import Logit
from ax.modelbridge.transforms.one_hot import OneHot
from ax.modelbridge.transforms.remove_fixed import RemoveFixed
from ax.modelbridge.transforms.standardize_y import StandardizeY
from ax.modelbridge.transforms.unit_x import UnitX
from botorch.models.transforms.input import InputTransform, Normalize
from botorch.models.transforms.outcome import OutcomeTransform, Standardize

TransformationTranslation: TypeAlias = Callable[
    [Any],  # NOTE: This should be `Transform`, but making types work in ax_to_botorch_transform_conversion is tricky.
    tuple[
        InputTransform | None,
        OutcomeTransform | None,
    ],
]


def ax_to_botorch_transform_input_output(
    transforms: list[Transform], outcome_names: list[str]
) -> tuple[InputTransform, OutcomeTransform]:
    """Determines the input and output transforms applied by Ax, and creates the equivalent transforms in botroch.

    This allows the botorch model internal to ax (which operates in a standard "model" space), to be used in the problem
    space (e.g non-standardised input and output). This is useful when calculating QoIs.

    Args:
        transforms: the TRAINED transforms that have been applied by `ax`.
            - Typically found at `TorchModelBridge.transforms`.
        outcome_names: the order of the output columns used to train the internal ax.Model.
            - Typically found at `TorchModelBridge.outcomes`.

    Return:
        - input_transform:
        - output_tranform:

    Using them in the following way allow input and output in the outcome/problem space:
        - Assume: `model` is a trained`botorch.models`(such as `TorchModelBridge.model.surrogate.model`)
        - > model.posterior(input_transform(X), posterior_transform = output_transform.untransform_posterior)

    Todo:
        - Ideally `ax_to_botorch_transform_conversion` would a config within the root of this module, so it could easily
            be exteneded. This is challenign becuase the 'translate_standardisey' function needs the specific
            outcome_names of the problem to be passed. This is because ax does not maintain the order of the metrics in
            the transform itself (it stores the names/order internally. See
            `ax.modelbridge.base.ModelBridge._transform_data` for details)
    """
    # This is a mapping from the ax Tranform to the function that translates it to botorch
    # Key: the specific transform class,
    # Value: the function to extract the input and outcome transforms
    #
    # NOTE: The transforms that are listed here are the standard transforms applied to model.BOTORCH_MODULAR.
    # They are detailed here:  ax.modelbridge.registry.MODEL_KEY_TO_MODEL_SETUP
    ax_to_botorch_transform_conversion: dict[type[Transform], TransformationTranslation] = {
        Cast: translate_cast,
        RemoveFixed: partial(check_transform_not_applied, parameter_names_store="fixed_parameters"),
        OrderedChoiceEncode: partial(check_transform_not_applied, parameter_names_store="encoded_parameters"),
        OneHot: partial(check_transform_not_applied, parameter_names_store="encoded_parameters"),
        IntToFloat: partial(check_transform_not_applied, parameter_names_store="transform_parameters"),
        Log: partial(check_transform_not_applied, parameter_names_store="transform_parameters"),
        Logit: partial(check_transform_not_applied, parameter_names_store="transform_parameters"),
        UnitX: translate_unitx,
        IVW: translate_ivw,
        Derelativize: translate_derelativize,
        # TODO(sw): Would be great to work out how to avoid passing col_names
        # Ax stores the order it uses here: `TorchModelBridge.outcomes`
        StandardizeY: partial(translate_standardisey, col_names=outcome_names),
    }

    input_transforms: list[InputTransform] = []
    output_transforms: list[OutcomeTransform] = []

    for t in transforms:
        try:
            converter = ax_to_botorch_transform_conversion[t.__class__]
        except KeyError:
            msg = f"ax_to_botorch_transform_conversion does not support {t.__class__}"
            raise NotImplementedError(msg) from None

        input_transform: InputTransform | None
        output_transform: OutcomeTransform | None
        input_transform, output_transform = converter(t)

        if input_transform is not None:
            input_transforms.append(input_transform)
        if output_transform is not None:
            output_transforms.append(output_transform)

    # TODO(sw 2024-11-21): Look at extending support to multiple transforms using:
    # botorch.models.transforms.outcome.ChainedOutcomeTransform
    # botorch.models.transforms.input.ChainedInputTransform
    if len(input_transforms) > 1:
        raise NotImplementedError(f"Currently only a single input transform is supported, found {input_transforms}")
    if len(output_transforms) > 1:
        raise NotImplementedError(f"Currently only a single output transform is supported, found {output_transforms}")

    return input_transforms[0], output_transforms[0]


#### Individual translation function


# TODO(sw): Maybe this is cleaner as an object rather than using functools?
def check_transform_not_applied(
    transform: Transform,
    parameter_names_store: str,
) -> tuple[InputTransform | None, OutcomeTransform | None]:
    """Used to ensure a transform has not been applied.

    Many transforms store an internal list of the ax.parameters (input) they should be applied to.
    This is determined by that parameter being of a specific type and having specific attributes as checked within the
    transform.
    This helper function is used to double check the transforms are not being used/applied to anything. This mean a
    translation from ax to botorch is not required.

    Args:
        transform: the transform to check
        parameter_names_store: The attribute on the transform that should be empty (falsey) if the tranform is not
            applied to anything.

    Returns:
        Input and output transforms required (will be None). Will raise an error if these transformation have actually
        been applied.

    Note:
        This should be instantiated with functools.partial, e.g.
            >>> from functools import partial
            >>> log_checker = partial(check_not_applied, parameter_names_store="transform_parameters")

    """
    # There are no names stored
    if not bool(getattr(transform, parameter_names_store)):
        return (None, None)

    msg = (
        f"expected {transform.__class__}.{parameter_names_store} to be empty/falsey,"
        f" instead found {getattr(transform,parameter_names_store)} indicating this transform is being used"
    )
    raise AssertionError(msg)


def translate_cast(transform: Cast) -> tuple[InputTransform | None, OutcomeTransform | None]:
    """Make sure that Cast has not flattned a HierachicalSearchSpace.

    Cast changes the parameter (e.g RangeParameter), castings the VALUE to the data type it should be.
        e.g RangeParameter values should be a float, cast the value to ensure it is a float
    It also deals with HierachicalSearchSpace:
        - (basically this is like a tree that navigates you to a more specific search space
            - e.g if 'parameter_a'> 2 -> use SearchSpace1
        - `.flatten_hss` flag if this has been used
    """
    if transform.flatten_hss:
        msg = "Ax transfrom to Botorch transform for Cast does not currently support HierachicalSearchSpace"
        raise NotImplementedError(msg)

    # No transform as it is assumed the inputs used for the botorch model are of they right type (e.g torch.tensor)
    return (None, None)


def translate_unitx(transform: UnitX) -> tuple[InputTransform | None, OutcomeTransform | None]:
    """Converts a trained UnitX to botorch equivalent.

    Ax bounds look like this: {'x1': (0.0, 2.0), 'x2': (0.0, 3.0)}

    BoTorch bounds look like: tensor([[0., 0.],[2., 3.]])
    """
    bounds = transform.bounds
    normalise = Normalize(d=len(bounds), bounds=torch.tensor(list(bounds.values())).T)
    _ = normalise.eval()
    return (normalise, None)


def translate_ivw(transform: IVW) -> tuple[InputTransform | None, OutcomeTransform | None]:  # noqa: ARG001
    """Handle IVW (inverse variance weight transform).

    IVW is used when at the same location (x), there are multiple measure of the same metric.
    It combines these into a single measure of the metric, and passes this on to the botroch model for training
    As this is only using for training (transforming the y input to the model), we can ignore this, as we currently
    use these transforms for prediction only.

    Note:
        It is hard to tell if this transformation has been applied because no attribute are stored on the object.

    Todo:
        Check if botorch supports multiple measure of a single metric at a single point (suspect not)
        - if not it is reasonable to ignore this transformation as standard botorch model shouldn't be using in that way
    """
    return (None, None)


def translate_derelativize(transform: Derelativize) -> tuple[InputTransform | None, OutcomeTransform | None]:  # noqa: ARG001
    """Handle Derelativize (relative constraints to non-relative constraints).

    Derelativize transforms optimisation configs and untransforms constraints. As we are only interested in input and
    output transformations, this can be ignored.

    Todo:
        Is there a safer way to ensure this is not being used? Difficult because it doesn't store anything internally

    Todo:
        This needs some additional work.
    """
    return (None, None)


def translate_standardisey(
    transform: StandardizeY,
    col_names: list[str],
) -> tuple[InputTransform | None, OutcomeTransform | None]:
    """Translate ax standardisation into botorch standardisation.

    Note:
        Ax does not maintain the order of the metrics, so need to explicitly pass the order.

        .. Todo::
            Can there be some work around for this? Would be good not to have to pass constraints.

    Note:
        col_name should be passed using functools.partial
            >>> from functools import partial
            >>> standardise_y = partial(translate_standardiseY, col_names=["loc", "scale"])

    Args:
        transform: StandardizeY to translate
        col_names: the order of the column in the data being passed in.
            This is required to the correct transformation can be applied to the correct column
    """
    # TODO(sw): this currently assumes all outputs are being standardises
    assert set(col_names) == set(transform.Ymean.keys())

    new_output_transform = Standardize(m=len(transform.Ymean))
    new_output_transform.means = torch.tensor([[transform.Ymean[name] for name in col_names]])
    new_output_transform.stdvs = torch.tensor([[transform.Ystd[name] for name in col_names]])
    # Need to set this directly because it is used to give the covariance
    new_output_transform._stdvs_sq = new_output_transform.stdvs.pow(2)  # noqa: SLF001
    new_output_transform._is_trained = torch.tensor([True])  # noqa: SLF001
    _ = new_output_transform.eval()

    return (None, new_output_transform)



"""Module for sampling from a posterior distribution."""

from .base import MeanVarPosteriorSampledEstimates, PosteriorSampler
from .independent_sampler import IndependentMCSampler
from .mean_sampler import MeanSampler
from .normal_independent_sampler import NormalIndependentSampler
from .ut_sampler import UTSampler



"""Contains samplers where the total dataset procuded is not effect by the size of the batch dimension used.

Batch invariance in 1d.
This can be achieved using the standard BatchSampler, for example:

    >>> from torch.utils.data import BatchSampler
    >>> data = [1, 2, 3, 4, 5]
    >>> list(BatchSampler(data, batch_size=3, drop_last=False))
    [1, 2, 3], [4, 5, 6]]
    >>> list(BatchSampler(data, batch_size=2, drop_last=False))
    [[1, 2], [3, 4], [5]]

Regardless of batch size, these results can be concatenated along the batch dimension to prodcude the same result.

Batch invariance in 2d:

    >>> data = [1, 2, 3, 4, 5, 6]
    We want to turn things into a 2d dataset, where the dimension being batched along (e.g rows) does not effect the
    final data produced. This will not be the case with MultiBatchSampler:

    >>> list(MultiBatchSampler(data, batch_shape=torch.Size([2, 3])))
    [[
        [1, 2, 3],
        [4, 5, 6]
    ]]
    >>> b1, b2, b3 = list(MultiBatchSampler(data, batch_shape=torch.Size([2, 1])))
    >>> print(f"{b1=},{b2=},{b3=})
    b1= [[1], b2= [[3], b3= [[5],
         [2]]      [4]]      [6]]
    >>>  concat([b1,b2,b3], axis = -1)
    [[ 1, 3, 5]
     [ 2, 4, 6]]

The datasets produced will put data in different location based on the batch shape.
This is aproblem if you will be aggrgating (e.g over rows), and were expecting the batches to be invariant over that
dimension.

Provide invariante batching in a specific dimension.
"""
# pyright: reportUnnecessaryTypeIgnoreComment=false

from collections.abc import Iterable, Iterator

import torch
from torch.utils.data import Sampler


class BatchInvariantSampler2d(Sampler[list[list[int]]]):
    """Returns 2d batchs where the final dataset in invariant to changes in the last batch dimension.

    The standard BatchSampler
        - has 1 row that it batches in sizes b.
        - returns items of shape (1) * b
    This BatchSampler has:
        - n rows that are batched in size b.
        - returns batches of n * b

    Importantly, the concatenated batches (along the last dimension) will alway the same 2d matrix,
    regardless of the bach size.

    Examples:
        >>> Conceptually the full dataset might have the following indexs
        [[ 1, 3, 5]
         [ 2, 4, 6]]

        >>> If batched along index = -1, with batch size 2 will return results as follows
        b1= [[1,3],   b2= [[5],
             [2,4]]]       [6]]

        >>> concat([b1, b2], axis=-1)
        [[ 1, 3, 5]
         [ 2, 4, 6]]

    This is important when the final dataset produced should be invariate to changes in b.

    Note:
        - The final batch can return a partial batch (n rows, less than b columns)
        - Related to issue #76

    Todo:
        This is a very specific case because its cater to general case.
            - Can we make a more general 2d case? The batched dim needs to be filled last.
              Any aggregation is then expected on the batch dim
            - Can we make a general higher dim case?
            - Should we just take multiple samplers and batch them in parallel?
                - Pro: Might be cleaner conceptually
                - Con: When we treat as a dataset we want to get through all data before we start repeating data
    """

    def __init__(self, sampler: Sampler[int] | Iterable[int], batch_shape: torch.Size) -> None:
        """Will produce batch (along the rows) of a 2d batch.

        Args:
            sampler (Sampler[int] | Iterable[int]): Sampler (e.g RandomSampler or SerquentialSampler) to be batched.
                - Must be Sized (e.g have __len__)
                - See torch DataLoader implmentation for an examples
            batch_shape (torch.Size): The batch shape created of the underlieing sample.
        """
        self.sampler = sampler

        if len(batch_shape) != 2:  # noqa: PLR2004
            msg = f"Only 2d batches are supported, but recieved {batch_shape=}"
            raise ValueError(msg)
        self.batch_shape = batch_shape

        # Upfront check the the batch size is correct
        # don't want to fail on the last batch after spending a lot of time training
        # NOTE: see note about calling len(sampler) in method __len__
        data_len: int = len(sampler)  # type: ignore[arg-type]
        partial_batch_data_len = data_len % batch_shape.numel()
        if partial_batch_data_len % batch_shape[0] != 0:
            msg = (
                f"Final batch will have {partial_batch_data_len} items,"
                f" which does not fit into batch shape {torch.Size([*batch_shape[:-1] , -1])}"
            )
            raise ValueError(msg)

    def __len__(self) -> int:  # noqa: D105
        # NOTE: len is not a gaurenteed method on Sampler. but all the subclasses we use have it
        #       see note about calling len(sampler) in method __len__
        data_len: int = len(self.sampler)  # type: ignore[arg-type]
        return (data_len + self.batch_shape.numel() - 1) // self.batch_shape.numel()

    def __iter__(self) -> Iterator[list[list[int]]]:
        """Construct the indexes that will be in the batch.

        Returns:
            Returns a nested lists of the shape specified by `batch_shape`.
        """
        # Do all batches except the last, which might be partial batch
        sample_iter = iter(self.sampler)
        for _ in range(len(self) - 1):
            # the follow with throw StopIteration if there is not enough data. This should flow to the user.
            _batch: list[int] = [next(sample_iter) for _ in range(self.batch_shape.numel())]
            # By default fills across the row, we need it to fill down the columns
            # Fill the transpose of the batch shape, then transpose it back
            batch: torch.Tensor = torch.tensor(_batch).view(self.batch_shape[::-1]).T
            yield batch.tolist()

        # determine the shape of the final batch
        batch = torch.tensor(list(sample_iter)).reshape(-1, self.batch_shape[0]).T
        yield batch.tolist()



import torch
from botorch.models.model import Model


def check_same_model_data(model_actual: Model, model_expected: Model):
    torch.testing.assert_close(model_actual.train_inputs[0], model_expected.train_inputs[0])
    torch.testing.assert_close(model_actual.train_targets, model_expected.train_targets)
    torch.testing.assert_close(model_actual.likelihood.noise_covar.noise, model_expected.likelihood.noise_covar.noise)


def check_posterior(model_actual: Model, model_expected: Model):
    torch.testing.assert_close(model_actual.distribution.mean, model_expected.distribution.mean)
    torch.testing.assert_close(
        model_actual.distribution.covariance_matrix, model_expected.distribution.covariance_matrix
    )



"""Module for quantity of interest Quantity of Interest estimator protocol."""

from typing import Protocol

import torch
from botorch.models.model import Model


class QoIEstimator(Protocol):
    """A protocol for quantity of interest (QoI) estimators."""

    def __call__(self, model: Model) -> torch.Tensor:
        """Function that produces multiple estimates of a scalar quantity of interest.

        If the function needs additional arguments,
        this should be handled by a class that implements this protocol or a closure.

        Args:
            model: A botorch model that is used to estimate the QoI.

        Returns:
            torch.Tensor: A tensor of QoI estimates with shape (number_of_estimates,).
        """
        ...

    def mean(self, x: torch.Tensor) -> torch.Tensor:
        """Function that computes the mean of the QoI estimates (the output of the call() method).

        For many applications this should just be using a default implementation that computes the mean.
        E.g using torch.mean(x).

        However, in some special cases, it might be useful to provide a custom implementation
        to give a more accurate estimate. E.g. when UTSampler is used.

        Args:
            x: A tensor of QoI estimates with shape (number_of_estimates,).

        Returns:
            torch.Tensor: The mean of the QoI estimates. Should be a scalar.
        """
        return torch.mean(x)

    def var(self, x: torch.Tensor) -> torch.Tensor:
        """Function that computes the variance of the QoI estimates (the output of the `call()` method).

        For many applications this should just be using a default implementation that computes the variance.
        E.g. using `torch.var(x)`.

        However, in some special cases, it might be useful to provide a custom implementation
        to give a more accurate estimate. E.g. when UTSampler is used.

        Args:
            x: A tensor of QoI estimates with shape (number_of_estimates,).

        Returns:
            torch.Tensor: The variance of the QoI estimates. Should be a scalar.
        """
        return torch.var(x)



"""Helper functions for ax Experiments."""

from collections.abc import Iterable, Mapping
from typing import Any, cast

import numpy as np
import pandas as pd
from ax import Data, Experiment, SearchSpace
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.types import TParameterization
from ax.modelbridge import Models
from numpy.typing import NDArray
from scipy.stats import rv_continuous

from axtreme.evaluation import EvaluationFunction
from axtreme.metrics import LocalMetadataMetric
from axtreme.runner import LocalMetadataRunner
from axtreme.simulator import Simulator
from axtreme.utils.distibution_helpers import distribution_parameter_names_from_scipy


def make_experiment(
    simulator: Simulator,
    search_space: SearchSpace,
    dist: rv_continuous,
    n_simulations_per_point: int = 100,
) -> Experiment:
    """Returns an experiment according to the given simulator, search_space, and dist_class.

    Args:
        simulator: The simulator to use for the experiment.
        search_space: The ax search space to use for the experiment.
        dist: The distribution that the result of a simulation is assumed to follow.
        n_simulations_per_point: The number of simulations to run for each point in the experiment.

    Returns:
        An ax Experiment.
    """
    # Define the metrics from the chosen distibution
    metrics = [
        LocalMetadataMetric(name=param, lower_is_better=False)
        for param in distribution_parameter_names_from_scipy(dist)
    ]
    # Optimisation config follows directly from the metrics
    mo = MultiObjective(
        objectives=[Objective(metric=m) for m in metrics],
    )
    optimization_config = MultiObjectiveOptimizationConfig(
        objective=mo,
        # TODO(sw): work out if we should provide these
        # Objective_threshold. They are only really needed when calulating the pareto_frontier.
    )

    # Construst runner config.
    # NOTE: This is completely determined by simulator and distibution
    # TODO(sw): issue #42 n_simulations_per_point: Need pick this is a more realistic way
    eval_function = EvaluationFunction(
        simulator=simulator,
        output_dist=dist,
        parameter_order=list(search_space.parameters.keys()),
        n_simulations_per_point=n_simulations_per_point,
    )
    runner = LocalMetadataRunner(eval_function)

    return Experiment(search_space=search_space, optimization_config=optimization_config, runner=runner)


def add_sobol_points_to_experiment(experiment: Experiment, n_iter: int = 5, seed: int | None = None) -> None:
    """Adds some points (chosen by sobol) to an experiment.

    Typically used to initialise the experiment so a GP has data for training.
    """
    # TODO(sw): this should deep copy the experiment,
    #           returning the new experiment rather than modifying an object in place?

    sobol = Models.SOBOL(search_space=experiment.search_space, seed=seed)
    for _ in range(n_iter):
        trial = experiment.new_trial(sobol.gen(1))
        _ = trial.run()
        _ = trial.mark_completed()


def add_metric_data_to_experiment(
    experiment: Experiment,
    parameterizations: Iterable[TParameterization],
    metric_data: Iterable[Mapping[str, float | tuple[float, float | None] | dict[str, float | None]]],
) -> tuple[Data, int]:
    """Add metric data to an experiment.

    This function is used to add data from previously run simulations to an experiment.
    To use this function one needs to have an estimate of the mean and standard error for each metric.
    This is useful if the simulator is slow or expensive to run and the data is already available.

    Args:
        experiment: The ax Experiment to add the data to.
        parameterizations: The parameterizations used in the simulation.

            - Iterable where each element is a dict parameter names and values ({[param_name]: value}).

        metric_data: The metric data from the simulation.

            - Iterable where each element is a dict of metric names and values.
            - The values can be a float, a tuple of (mean, sem), or a dict of {"mean": mean, "sem": sem}.
            - If a float is provided, the sem is assumed to be 0.

    Returns:
        - The data that was attached to the experiment.
        - The trial index of the trial that was created.
    """
    # Attach all the parameterizations to the trial (one arm per parameterization)

    # First check that all parameterizations are in the search space
    # Only attach the parameterizations that are in the search space
    # Only use the metric data for the parameterizations that are in the search space
    validated_parameterizations: list[TParameterization] = []
    validated_metric_data: list[Mapping[str, float | tuple[float, float | None] | Mapping[str, float | None]]] = []
    for parameterization, metric_data_for_param in zip(parameterizations, metric_data, strict=False):
        try:
            experiment.search_space.validate_membership(parameterization)
            validated_parameterizations.append(parameterization)
            validated_metric_data.append(metric_data_for_param)
        except ValueError as e:
            msg = f"Parameterization {parameterization} is not in the search space.\n\t{e!s}"
            print(msg)  # noqa: T201

    arms, trial_index = experiment.attach_trial(parameterizations=validated_parameterizations)

    # Creating a pandas DataFrame to store the output data for all the arms
    rows = []
    for arm_name, metric_result in zip(arms, validated_metric_data, strict=False):
        # Adding a row for each metric
        for metric_name in experiment.metrics:
            # Get mean and sem from the distribution result
            metric_mean_sem = metric_result[metric_name]
            mean: float | None = None
            sem: float | None = None
            if isinstance(metric_mean_sem, dict):
                mean = metric_mean_sem["mean"]
                sem = metric_mean_sem["sem"]
            elif isinstance(metric_mean_sem, tuple):
                mean, sem = metric_mean_sem
            elif isinstance(metric_mean_sem, float):
                mean = metric_mean_sem
            else:
                msg = f"Expected metric_data to be of type float, tuple or dict, got: {type(metric_mean_sem)}"
                raise TypeError(msg)

            assert mean is not None, f"Expected mean of metric {metric_name} of {arm_name} to be a float, got None"
            # Append the resulting new row
            new_row = {
                "arm_name": arm_name,
                "metric_name": metric_name,
                "mean": mean,
                "sem": sem,
                "trial_index": trial_index,
            }
            rows.append(new_row)

    results_data = pd.DataFrame(rows)
    data = Data(df=results_data)
    # Attach the data to the experiment without overwriting the previous data
    _ = experiment.attach_data(data=data, combine_with_last_data=True)

    # Marking the trial as completed so that it can be used for training the model
    trials = experiment.get_trials_by_indices(trial_indices=[trial_index])
    _ = trials[0].mark_completed()

    # Return the data that was attached
    return data, trial_index


def add_simulation_data_to_experiment(
    experiment: Experiment,
    parameters: list[str],
    simulation_inputs: Iterable[NDArray[np.float64]],
    simulation_outputs: Iterable[NDArray[np.float64]],
) -> tuple[Data, int]:
    """Add raw simulator data to an experiment.

    This function is used to add data from previously run simulations to an experiment.
    This is very useful if the simulator is slow or expensive to run and the data is already available.
    Using this function allows you to add the data to the experiment and then use it to train the model
    which can be used to collect new data by running new simulations.

    Args:
        experiment: The ax Experiment to add the data to.
        parameters: The names of the parameters used as input to the simulator.

            - This has to match the names of the parameters in the search space of the experiment.
            - This should match the order of the input dimensions in the simulation_inputs.

        simulation_inputs: The inputs to the simulator.

            - Iterable where each element is the parameters used in the simulation.
            - Each element should be a numpy array with shape (n_input_dims,)

        simulation_outputs: The outputs from the simulator.

            - Iterable where each element is multiple output from the simulation ran with the same input.
            - Each element should be a numpy array with shape (n_simulations_per_point,)
            - Only supports a single output dimension for now.

    Returns:
        - The data that was attached to the experiment.
        - The trial index of the trial that was created.
    """
    _: Any
    runner = experiment.runner
    assert isinstance(
        runner, LocalMetadataRunner
    ), f"Expected experiment.runner to be of type LocalMetadataRunner, got: {type(runner)}"

    # Change format of inputs into parameterization dicts
    parameterizations: list[TParameterization] = []
    for x in simulation_inputs:
        flat_parameters = x.flatten()

        parameterizations.append({parameter_name: flat_parameters[i] for i, parameter_name in enumerate(parameters)})

    # Use the post processing from the evaluation function to get
    # the distribution parameters from the simulation outputs
    simulation_distribution_results = [
        runner.evaluation_function.post_process_simulation_output(y) for y in simulation_outputs
    ]

    # Fetch the mean and sem for each metric from the simulation distribution results
    distribution_metric_results: list[dict[str, float | tuple[float, float | None] | dict[str, float | None]]] = [
        {metric_name: distribution_result.metric_data(metric_name) for metric_name in experiment.metrics}
        for distribution_result in simulation_distribution_results
    ]

    # Use the helper function to add the metric data to the experiment
    return add_metric_data_to_experiment(experiment, parameterizations, distribution_metric_results)


def extract_data_from_experiment_as_json(
    experiment: Experiment,
) -> dict[int, dict[str, dict[str, dict[str, float | dict[str, float]]]]]:
    """Extracts the data from an ax Experiment.

    Args:
        experiment: The Ax experiment to extract the data from.

    Returns:
        A dictionary with the following structure:

        .. code-block:: json

            {
                "trial_index": {
                    "arm_name": {
                        "parameters": {
                            "parameter_name": "parameter_value"
                        },
                        "metrics": {
                            "metric_name": {
                                "mean": 0.0,
                                "sem": 0.0
                            }
                        }
                    }
                }
            }
    """
    metric_data = experiment.fetch_data().df
    trial_indexes = metric_data[["trial_index"]].drop_duplicates()
    trials = experiment.get_trials_by_indices(trial_indexes["trial_index"].values)

    trial_data: dict[int, dict[str, dict[str, dict[str, float | dict[str, float]]]]] = {}

    metric_data_by_arm = metric_data.groupby("arm_name")

    for trial in trials:
        if trial.index not in trial_data:
            trial_data[trial.index] = {}
        for arm in trial.arms:
            if arm.name not in trial_data[trial.index]:
                trial_data[trial.index][arm.name] = {
                    "parameters": {},
                    "metrics": {},
                }

            for parameter_name, parameter_value in arm.parameters.items():
                trial_data[trial.index][arm.name]["parameters"][parameter_name] = parameter_value

            arm_metric_data = metric_data_by_arm.get_group(arm.name)[["metric_name", "mean", "sem"]]
            for metric_name, mean, sem in arm_metric_data.to_numpy():
                trial_data[trial.index][arm.name]["metrics"][metric_name] = {"mean": mean, "sem": sem}

    return trial_data


def add_json_data_to_experiment(
    experiment: Experiment, json_data: dict[int, dict[str, dict[str, dict[str, float | dict[str, float]]]]]
) -> None:
    """Adds the data from a dictionary to an ax Experiment.

    Args:
        experiment: The ax Experiment to add the data to.
        json_data: The data to add to the ax Experiment. The structure should be the same as the output of
            extract_data_from_experiment.

    Example:
        .. code-block:: json

            {
                "trial_index": {
                    "arm_name": {
                        "parameters": {
                            "parameter_name": "parameter_value"
                        },
                        "metrics": {
                            "metric_name": {
                                "mean": 0.0,
                                "sem": 0.0
                            }
                        }
                    }
                }
            }
    """
    # For each arm in the dictionary, extract the parameters and metrics for all arms in the trial
    for trial_arms in json_data.values():
        # The parameters should always be of type dict[str, float]
        # So we use cast to tell mypy that we know this is the case
        trial_parameters: list[dict[str, float]] = [
            cast(dict[str, float], arm_data["parameters"]) for arm_data in trial_arms.values()
        ]

        # The metrics should always be of type dict[str, dict[str, float]]
        # So we use cast to tell mypy that we know this is the case
        trial_metrics: list[dict[str, dict[str, float | None]]] = [
            cast(dict[str, dict[str, float | None]], arm_data["metrics"]) for arm_data in trial_arms.values()
        ]
        _ = add_metric_data_to_experiment(
            experiment=experiment,
            parameterizations=trial_parameters,
            metric_data=trial_metrics,
        )



"""Module for the UTSampler class for sampling a posterior using the Unscented Transform."""

import numpy as np
import torch
from botorch.posteriors import Posterior
from filterpy.kalman import MerweScaledSigmaPoints

from axtreme.sampling.independent_sampler import IndependentMCSampler


class UTSampler(IndependentMCSampler):
    """Class for sampling a posterior using the Unscented Transform.

    The Unscented Transform is a method for transforming a distribution through a non-linear function.
    The method is based on the Unscented Kalman Filter.
    It uses a set of sigma points to estimate the mean and covariance of the transformed distribution.
    """

    def __init__(
        self,
        alpha: float = 1,
        beta: float = 2,
        kappa_base: float = 3,
    ) -> None:
        """Initializer for the UTSampler.

        Args:
            alpha: Scaling parameter for the sigma points. See filterpy.kalman.MerweScaledSigmaPoints.
            beta: Parameter for the distribution.
                - For Gaussian distributions, beta=2 is optimal.
                - See filterpy.kalman.MerweScaledSigmaPoints.
            kappa_base: Base parameter for the sigma points. See filterpy.kalman.MerweScaledSigmaPoints.
        """
        # We don't want to use the .__init__() on MCSampler as to forces `sample_shape` which is currently unknown.
        # Instead we use nn.Module (root class) because it performs internal set up.
        torch.nn.Module.__init__(self)

        # This is part of the design contact specific by MCSampler. We want to conform but it irrelevant for this class.
        self.seed = None

        # This is part of the design contact specific by MCSampler. We want to conform but need to create it later
        self.sample_shape: torch.Size | None = None

        # Make these private. Users should create a new sampler if they want different parameters
        self._alpha = alpha
        self._beta = beta
        self._kappa_base = kappa_base

        self.register_buffer("base_samples", None)

    @property
    def alpha(self) -> float:
        """The alpha parameter defined by filterpy.kalman.MerweScaledSigmaPoints."""
        return self._alpha

    @property
    def beta(self) -> float:
        """The beta parameter defined by filterpy.kalman.MerweScaledSigmaPoints."""
        return self._beta

    @property
    def kappa_base(self) -> float:
        """The kappa_base parameter related to kappa defined by filterpy.kalman.MerweScaledSigmaPoints.

        The kappa value used in filterpy.kalman.MerweScaledSigmaPoints is calculated as `kappa_base - num_target`.
        """
        return self._kappa_base

    def _construct_base_samples(self, posterior: Posterior) -> None:
        """Build the base samples that should be used for sampling the GP output produced by a single input (x).

        `IndependentMCSampler`s apply the same base samples to every x in the posterior (independantly). The base
        samples to be used are define in this method.

        Args:
            posterior: the posterior to construct the base samples for.

        Return:
            None. Stores base_samples of shape (*b,n,m) in registered buffer.
        """
        # has shape (*b,n,m)
        explicit_posterior_shape = self._explicit_posterior_shape(posterior)
        # This is the t dim
        target_dim = explicit_posterior_shape[-1]

        if (self.base_samples is None) or (self.base_samples.shape[-1] != target_dim):
            points, mean_weights, cov_weights = calculate_sigmas(target_dim, self.alpha, self.beta, self.kappa_base)

            self.sample_shape = torch.Size([points.shape[0]])

            self.register_buffer("base_samples", points)
            self.register_buffer("mean_weights", mean_weights)
            self.register_buffer("cov_weights", cov_weights)

            for attr_name in ["base_samples", "mean_weights", "cov_weights"]:
                attr = getattr(self, attr_name)
                if attr.device != posterior.device:
                    _ = self.to(device=posterior.device)
                if attr.dtype != posterior.dtype:
                    _ = self.to(dtype=posterior.dtype)

    def mean(self, transformed_points: torch.Tensor, dim: int = 0) -> torch.Tensor:
        """Estimate the mean of the transformed UT sampled points.

        The UT sampled points are designed together with the weights to estimate the mean of the distribution
        When the points are transformed through some function,
        the mean of the transformed points is estimated by calculating the weighted sum of the points.

        This will only work correctly if the last points generated using this sampler
         - Have the same number of targets (m) the transformed points.
         - Were generated using the same alpha, beta, and kappa_base parameters.

        Args:
            transformed_points: The transformed points to estimate the mean of.
            dim: The dimension along which to calculate the mean.

        Returns:
            The estimated mean of the distribution.
        """
        if transformed_points.shape[dim] != self.mean_weights.shape[0]:
            msg = f"Expected {self.mean_weights.shape[0]} points, got {transformed_points.shape[0]}"
            raise ValueError(msg)

        return torch.tensordot(self.mean_weights[None], transformed_points.moveaxis(dim, 0), dims=1).squeeze(0)

    def var(self, transformed_points: torch.Tensor, dim: int = 0) -> torch.Tensor:
        """Estimate the variance of the transformed UT sampled points.

        The UT sampled points are designed together with the weights to estimate the variance of the distribution
        When the points are transformed through some function,
        the variance of the transformed points is estimated by calculating the weighted sum of the points.

        This will only work correctly if the last points generated using this sampler
         - Have the same number of targets (m) as the transformed points.
         - Were generated using the same alpha, beta, and kappa_base parameters.

        Args:
            transformed_points: The transformed points to estimate the variance of.
            dim: The dimension along which to calculate the variance.

        Returns:
            The estimated variance of the distribution.
        """
        moved_axis_points = transformed_points.moveaxis(dim, 0)

        if moved_axis_points.shape[0] != self.mean_weights.shape[0]:
            msg = f"Expected {self.mean_weights.shape[0]} points, got {moved_axis_points.shape[0]}"
            raise ValueError(msg)

        if moved_axis_points.shape[0] != self.cov_weights.shape[0]:
            msg = f"Expected {self.cov_weights.shape[0]} points, got {moved_axis_points.shape[0]}"
            raise ValueError(msg)

        centered_points = moved_axis_points - self.mean(moved_axis_points, dim=0)[None]
        return torch.tensordot(self.cov_weights[None], torch.square(centered_points), dims=1).squeeze(0)


# This function is not part of the class because it is not specific to the class.
# However, it could be moved to the class if that is deemed more appropriate.
def calculate_sigmas(
    dim: int, alpha: float, beta: float, kappa_base: float
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Create the sigma points, and the weight for calculating mean and variance.

    Args:
        dim: dimensionality of the space to produce sigma points for.
            E.g The dimensionality of the output of a GP at a sinlge input point.
        alpha: Scaling parameter for the sigma points.
        beta: Parameter for the distribution. For Gaussian distributions, beta=2 is optimal.
        kappa_base: Base parameter for the sigma points.

    Return:
        A tuple of three tensors where,

        1. (m,dim) defines signma points, where
           - m: is the number of sigma points generated
           - dim: is the dimension of those points
        2. (m): weights required to combines these signma points for a mean estimate
        3. (m): weights required to combines these signma points for a variance estimate

    Todo:
        - does not currently support correlation between t dimensions.
        - (ks) - For now using filterpy for UT. Might want to implement our own version
    """
    sigmas = MerweScaledSigmaPoints(n=dim, alpha=alpha, beta=beta, kappa=kappa_base - dim)

    points = torch.tensor(sigmas.sigma_points(np.zeros(dim), np.eye(dim)))
    mean_weights = torch.tensor(sigmas.Wm)
    cov_weights = torch.tensor(sigmas.Wc)

    return points, mean_weights, cov_weights



from .batch_invariant_sampler import BatchInvariantSampler2d
from .dataset import MinimalDataset
from .importance_dataset import ImportanceAddedWrapper, ImportanceIndexWrapper
from .multi_dim_batch_sampler import MultiBatchSampler
from .numpy_file_dataset import NumpyFileDataset
from .sizable_sequential_sample import SizableSequentialSampler
from .fixed_random_sample import FixedRandomSampler



"""Functions to configure logging for the application."""

import logging
import sys
from pathlib import Path

__all__ = ["configure_logging"]

logger = logging.getLogger(__name__)


def configure_logging(
    log_level_console: str = "WARNING",
    log_file: Path | None = None,
    log_level_file: str = "WARNING",
) -> None:
    """Configure logging for the application, allowing for both console and file logging.

    Sets the log levels and formats for the output, ensuring that logs are captured as specified.

    Parameters
    ----------
    log_level_console : str, optional
        log level for console output, by default "WARNING"
    log_file : Path | None, optional
        log file to be used. If None, file logging is disabled. by default None
    log_level_file : str, optional
        log level for file output, by default "WARNING"

    Raises:
    ------
    TypeError
        if an invalid value for log_level_console or log_level_file is passed

    Examples:
    --------
        configure_logging(log_level_console="INFO", log_file=Path("app.log"), log_level_file="DEBUG")
    """
    # sourcery skip: extract-duplicate-method, extract-method
    log_level_console_numeric = getattr(logging, log_level_console.upper(), None)
    if not isinstance(log_level_console_numeric, int):
        raise TypeError(f"Invalid log level to console: {log_level_console_numeric}")

    log_level_file_numeric = getattr(logging, log_level_file.upper(), None)
    if not isinstance(log_level_file_numeric, int):
        raise TypeError(f"Invalid log level to file: {log_level_file_numeric}")

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level_console_numeric)
    console_formatter = logging.Formatter("%(levelname)-8s %(message)s")
    console_handler.setFormatter(console_formatter)
    root_logger.addHandler(console_handler)

    if log_file:
        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(str(log_file.absolute()), "a")
        print(f"Logging to: {log_file.absolute()}")  # noqa: T201
        file_handler.setLevel(log_level_file_numeric)
        file_formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s", "%Y-%m-%d %H:%M:%S")
        file_handler.setFormatter(file_formatter)
        root_logger.addHandler(file_handler)

    return



"""Helper function for plotting DoE behaviour."""

from typing import Any, TypeAlias

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

Numpy2dArray: TypeAlias = np.ndarray[tuple[int, int], np.dtype[np.float64]]


def plot_qoi_estimates(
    results: Numpy2dArray,
    ax: None | Axes = None,
    q: tuple[float, ...] = (0.1, 0.5, 0.9),
    points_between_ests: int = 1,
    name: str | None = None,
    **kwargs: Any,  # noqa: ANN401
) -> Axes:
    """Plots how the QoI estimates changes over DoE process.

    Args:
        results: shape (n_doe_rounds, n_qoi_estimates).

            - n_doe_rounds: The number of DoE rounds in which a QoI estimate was produced.
            - n_qoi_estimates: the number of estimates produced by a single run of the QoI estimator.

        ax: ax to add the plots to. If not provided, one will be created.
        q: the quantiles that should be used/reported.
        points_between_ests: This should be used if multplie DoE iterations are used between qoi estimates
            (e.g if the estimate is expensive). It adjusts the scale of the x axis.
        name: optional name that should be added to the legend information for this plot
        kwargs: kwargs that should be passed to matplotlib. Must be applicable to `ax.plot` and `ax.fill_between`

    Returns:
        Axes: the ax with the plot.
    """
    quantiles = np.quantile(results, q=q, axis=1)

    if ax is None:
        _, ax = plt.subplots()

    x = range(1, (len(results) + 1) * points_between_ests, points_between_ests)
    _ = ax.fill_between(
        x,
        quantiles[0],
        quantiles[-1],
        label=f"{q[0] * 100}% to {q[-1] * 100}% Confidence Bound {name}",
        alpha=0.3,
        **kwargs,
    )

    for q_idx in range(1, len(q) - 1):
        _ = ax.plot(x, quantiles[q_idx], **kwargs)

    return ax



"""Helpers for understanding the population values expected by estimators.

NOTE: These tool provide indicative/approximate result.
"""

from typing import Any

import matplotlib.pyplot as plt
import torch
from matplotlib.axes import Axes
from scipy.stats import gaussian_kde
from torch.distributions import Distribution, Normal, StudentT


def sample_mean_se(samples: torch.Tensor) -> StudentT:
    """Distibution of the population mean as estimated by this sample.

    Note:
        The distribution of the sample itself doesn't matter. The output distibution is not effected by this.

    Use the following `link <https://stats.libretexts.org/Courses/Lake_Tahoe_Community_College/Book%3A_Introductory_Statistics_(OpenStax)_With_Multimedia_and_Interactivity_LibreTexts_Calculator/08%3A_Confidence_Intervals/8.03%3A_A_Single_Population_Mean_using_the_Student_t-Distribution>`_

    se = sigma / n**.5

        - Sigma: should be the population standard deviation, but we approximate this with the sample standard deviation
        - Because of this approximation we use the Student-t distribution

    Args:
        samples: 1d tensor of sample to estimate the population mean from

    Return:
        Distribution of the population mean based on the provided sample. Additionally, it provides the 95% confidence
        bounds for the estimate, which are typically calculated to fall within a range of 93% to 97% coverage, depending
        on sample variability and the assumptions of the calculation method.

    Todo:
        - .cdf() raises NotImplementedError for torch implemenation of StudentT. This is annoying
            because this is the best way to check the confidence bounds (using z = (y - mean)/stddev) assumes you
            are using a normal distibution rather than a student t distibution. This approximation is considered okay
            for n>30)
    """
    sample_mean = samples.mean()
    # Note by default torch.std() give the sample std: 1/(n-1) * sum[(x_i - x_bar)**2]
    sample_mean_se = samples.std() / len(samples) ** 0.5
    return StudentT(df=len(samples) - 1, loc=sample_mean, scale=sample_mean_se)


def sample_median_se(samples: torch.Tensor) -> Normal:
    """Distibution of the population median as estimated by this sample.

    Details of this method can be found `here. <https://en.wikipedia.org/wiki/Median#Sampling_distribution.>`_

    Note:
        This function relies on the approximation `estimate_pdf_value_from_sample`. The approximation can be quite
        inaccurate (see the function for details), and as a result this function should be treated as an estimate.

    Note:
        The result returned are much more noisey than sample_mean_se.

    Args:
        samples: 1d tensor of sample to estimate the population median from

    Return:
        Distibution of the population median as estimated by this sample. The 95% bounds (with 50 samples) estimated by
        this function typically produce bounds actually between 90%-97%.
    """
    sample_median = samples.median()
    # pdf at median point
    f_m = estimate_pdf_value_from_sample(samples, float(sample_median))

    sample_median_se = (1 / (4 * len(samples) * f_m**2)) ** 0.5
    return Normal(loc=sample_median, scale=sample_median_se)


def estimate_pdf_value_from_sample(sample: torch.Tensor, x: float) -> float:
    """Construct a distibution from a sample, and get the pdf value at point x.

    WARNING: This is an approximate method, and results impove with more samples. See testing results below.

    Args:
        sample: 1d Samples to construct a pdf from.
        x: the point at which to evaluate the pdf.

    Return:
        Estimated pdf value.

    Testing results:
    The mean returned and the cof have the following behaviour. Full test detail can be run at
    `tests/utils/test_population_estimators.py : visualise_performance_of_estimate_pdf_value_from_sample`

    Number of samples | mean_est/true |   Coef   |
    ==============================================
    11                |  .86  - 1.02  | .25-.30  |
    22                |  .88  - 1.02  | .18-.20  |
    44                |  .90  - 1.01  | .14-.16  |
    88                |  .92  - 1.00  | .11-.13  |
    176               |  .95  - 1.00  | .05-.10  |
    """
    kde = gaussian_kde(sample.numpy())
    return float(kde(x)[0])


def plot_dist(
    dist: Distribution,
    confidence_level: float = 3.0,
    ax: Axes | None = None,
    **kwargs: Any,  # noqa: ANN401
) -> Axes:
    """Plot the distribution PDF over domain `mean +- confidence_level * std`.

    Args:
        dist: the distribution to plot the pdf of.
        confidence_level: controls the width of the plot
        ax: the axes to plot on. If None, will create an x
        **kwargs: passed to the plotting method.

    Return:
        Axes with the plot.
    """
    if ax is None:
        ax = plt.subplot()

    x = torch.linspace(dist.mean - confidence_level * dist.stddev, dist.mean + confidence_level * dist.stddev, 100)

    _ = ax.plot(x, dist.log_prob(x).exp(), **kwargs)

    return ax



"""A variety of different types of models.

Key attributes of the model change dimension based on:
- the number of outputs (m)
- the type of noise

It is invariant to:
- (n): the number of input datapoints
- (d): the dimensionality of input datapoints

NOTE: these are not fixture as then you can not use them in parameterise
NOTE: This could be within conftest, but appears it is considerd bad practice to import from conftest https://github.com/pytest-dev/pytest/issues/3272
"""

import torch
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.transforms.outcome import Standardize

torch.set_default_dtype(torch.float64)


def get_train_x() -> torch.Tensor:
    """X dimensions are not effected by other setting, so one is used for all veriations."""
    gen = torch.Generator()
    _ = gen.manual_seed(7)
    return torch.rand(10, 1, dtype=torch.float64, generator=gen)


def single_task_homo_noise_m_1():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit

    `model.train_inputs`: ((n,d),)
    `model.train_targets`: (n,)
    `model.likelihood.noise_covar.noise`: (1,)
    """
    # (n,1)
    train_x = get_train_x()
    # (n,1)
    train_y = torch.sin(train_x)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=1)
    std_y = outcome_transform(train_y)[0]
    return SingleTaskGP(train_x, std_y)


def single_task_homo_noise_m_2():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit

    `model.train_inputs`: ((m,n,d),)
    `model.train_targets`: (m,n)
    `model.likelihood.noise_covar.noise`: (m,1)
    """
    # (n,1)
    train_x = get_train_x()  # (n,1)
    # (n,m = 2)
    train_y = torch.concat([torch.sin(train_x), torch.cos(train_x)], dim=-1)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=2)
    std_y = outcome_transform(train_y)[0]
    return SingleTaskGP(train_x, std_y)


def single_task_fixed_noise_m_1():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit

    `model.train_inputs`: ((n,d),)
    `model.train_targets`: (n,)
    `model.likelihood.noise_covar.noise`: (n,)
    """
    # (n,1)
    train_x = get_train_x()
    # (n,1)
    train_y = torch.sin(train_x)
    train_yvar = torch.linspace(0.01, 0.2, train_y.numel()).reshape_as(train_y)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=1)
    std_y, std_var = outcome_transform(train_y, train_yvar)
    return SingleTaskGP(train_x, std_y, std_var)


def single_task_fixed_noise_m_2():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit

    `model.train_inputs`: ((m,n,d),)
    `model.train_targets`: (m,n)
    `model.likelihood.noise_covar.noise`: (m,n)
    """
    # (n,1)
    train_x = get_train_x()
    # (n,m = 2)
    train_y = torch.concat([torch.sin(train_x), torch.cos(train_x)], dim=-1)
    train_yvar = torch.linspace(0.01, 0.2, train_y.numel()).reshape_as(train_y)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=2)
    std_y, std_var = outcome_transform(train_y, train_yvar)
    return SingleTaskGP(train_x, std_y, std_var)


def single_task_homo_noise_m_2_b3():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit.

    `model.train_inputs`: ((b,m,n,d),)
    `model.train_targets`: (b,m,n)
    `model.likelihood.noise_covar.noise`: (b,m,1)
    """
    # (n,1)
    train_x = get_train_x()  # (n,1)
    train_x = train_x.expand(3, *train_x.shape)
    # (b =3, n = 10,m = 2)
    train_y = torch.concat([torch.sin(train_x), torch.cos(train_x)], dim=-1)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=2, batch_shape=torch.Size([3]))
    std_y = outcome_transform(train_y)[0]
    return SingleTaskGP(train_x, std_y)


def single_task_homo_noise_m_1_b3():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit.

    `model.train_inputs`: ((b,n,d),)
    `model.train_targets`: (b,n)
    `model.likelihood.noise_covar.noise`: (b,1)
    """
    # (n,1)
    train_x = get_train_x()  # (n,1)
    train_x = train_x.expand(3, *train_x.shape)
    # (b =3, n = 10,m = 1)
    train_y = torch.sin(train_x)
    # Stops it complaining about non-centred input
    outcome_transform = Standardize(m=1, batch_shape=torch.Size([3]))
    std_y = outcome_transform(train_y)[0]
    return SingleTaskGP(train_x, std_y)


def single_task_fixed_noise_m_1_outcome_transform():
    """Produce the following shapes for its attributes.

    NOTE: hyperparameters of this model have not been fit

    `model.train_inputs`: ((n,d),)
    `model.train_targets`: (n,)
    `model.likelihood.noise_covar.noise`: (n,)
    """
    # (n,1)
    train_x = get_train_x()
    # (n,1)
    train_y = torch.sin(train_x)
    train_yvar = torch.linspace(0.01, 0.2, train_y.numel()).reshape_as(train_y)
    return SingleTaskGP(train_x, train_y, train_yvar, outcome_transform=Standardize(m=1))



"""Classes to organize the evaluation of a QoIs."""

# pyright: reportUnnecessaryTypeIgnoreComment=false

# %%
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import torch
from botorch.models.model import Model
from dataclasses_json import config, dataclass_json

import axtreme.eval.object_logging
from axtreme.eval import utils
from axtreme.qoi import QoIEstimator


@dataclass_json
@dataclass
class QoIJobResult:
    """Dataclass to store the results of running a QoIEstimator (as produced by QoIJob).

    Args:
        mean: the mean of the QoIEstimator results.
            Note, if UT methods are used in the QoIEstimator, this can't be calculated directory form samples
        var: variance in the QoIEstimator results.
            Note, if UT methods are used in the QoIEstimator, this can't be calculated directory form samples
        samples: Samples produced by the QoIEstimator
        tags: Label assigned to the result by the user. This information can typically be found somewhere in metadata,
            putting them here is for convience (more accessible and less noise).
        metadata: Optional indepth information about the conditions that produced these results.
    """

    # 0d tensors
    mean: torch.Tensor = field(metadata=config(encoder=torch.Tensor.tolist, decoder=torch.tensor))
    var: torch.Tensor = field(metadata=config(encoder=torch.Tensor.tolist, decoder=torch.tensor))
    # all samples
    samples: torch.Tensor = field(metadata=config(encoder=torch.Tensor.tolist, decoder=torch.tensor))

    tags: dict[str, str | float] = field(default_factory=dict)

    metadata: dict[str, Any] = field(default_factory=dict)

    # Implementation from dataclasses_json. Method stub to assist type checkers
    def to_dict(self) -> dict[str, Any]: ...  # type: ignore[empty-body]  # noqa: D102

    def from_dict(self) -> "QoIJobResult": ...  # type: ignore[empty-body]  # noqa: D102


@dataclass
class QoIJob:
    """Helper to make the interface below clear."""

    model: Model
    qoi: QoIEstimator
    name: str | None = None
    # easy access to key that will be used to search the job
    tags: dict[str, str | float] = field(default_factory=dict)
    # general metadata
    metadata: dict[str, str | float] = field(default_factory=dict)

    def __call__(self, output_file: None | Path = None) -> QoIJobResult:
        """Run the QoIJob.

        Args:
            output_file: If provided a json version of the results is appended to this file. Expects the file to be a
            list object. The file will be created if it does not exist.

        Returns:
            QoIJobResult
        """
        start_time = time.time()
        results = self.qoi(self.model)
        duration = time.time() - start_time

        mean = self.qoi.mean(results)
        var = self.qoi.var(results)
        metadata = {
            **self.metadata,
            "runtime_sec": duration,
            "qoi": axtreme.eval.object_logging.unpack_object_str_content(self.qoi, depth=2),
            "model": axtreme.eval.object_logging.unpack_object_str_content(self.model),
        }

        result = QoIJobResult(mean=mean, var=var, samples=results, tags=self.tags, metadata=metadata)

        if output_file:
            utils.append_to_json(result.to_dict(), output_file)

        return result



"""QoILookAhead acquisition function that looks ahead at possible models and optimize according to a quantity of interest."""  # noqa: E501

import warnings
from contextlib import ExitStack
from typing import Any

import torch
from ax.models.torch.botorch_modular.optimizer_argparse import _argparse_base, optimizer_argparse
from botorch.acquisition import AcquisitionFunction
from botorch.acquisition.input_constructors import acqf_input_constructor
from botorch.models import SingleTaskGP
from botorch.models.model import Model
from botorch.utils.transforms import t_batch_mode_transform
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood, GaussianLikelihood

from axtreme.qoi.qoi_estimator import QoIEstimator
from axtreme.sampling import MeanSampler, PosteriorSampler
from axtreme.utils.model_helpers import get_target_noise_singletaskgp, get_training_data_singletaskgp


class QoILookAhead(AcquisitionFunction):
    """QoILookAhead is a generic acquisition function that estimates the usefulness of a design point.

    It estimates the usefulness of a design point by:

    - Creating a new model(s) (e.g the looking ahead) by including the design point in the training data.

      .. Note::
        The new model condition on the design point, it does not update hyperparameters.

    - Calculate the QoI with the new model, and find the variance of this distribution.

    The design point that results in the lowest variance QoI is considered the most desireable.

    Note on Optimisation:
        Optimising the AcquisitionFunction is an important part of the DoE process. The stochasticity and smoothness of
        the acquisition function determine what optimisers can be used. This acquisition function has the following
        properties with the default setup:

    Smoothness:
        QoILookAhead is smooth (twice differentiable) if:

        - The model used for instantiation produces smooth outputs.
        - A sampler produces smooth y values.
        - The method used to produce y_var estimates is smooth.
        - The QoIEstimator is smooth with respect to the model (e.g small changes in the model produce smooth
          change in the QoIEstimator result.)

        The default optimiser assume the QoI may not be smooth, and uses a gradient free optimiser. If your QoI is
        smooth these setting should be overridden.

    Stochasticity:
        QoILookAhead is deterministic if all the above components are deterministic.
        With default settings it is deterministic.
    """

    def __init__(
        self,
        model: SingleTaskGP,
        qoi_estimator: QoIEstimator,
        sampler: PosteriorSampler | None = None,
    ) -> None:
        """QoILookAhead acquisition function.

        Args:
            model: A fitted model. Only SingleTaskGP models are currently supported.

                - General GpytorchModel may eventually be supported.

            qoi_estimator: A callable conforming to QoIEstimator protocol.
            sampler: A sampler that is used to sample fantasy observations for each candidate point.
                If None, a MeanSampler is used. This then uses the mean of the posterior as the fantasy observation.

                .. Note::
                    Sampler choice can effect the stochasticty and smoothness of the acquisition function. See class
                    docs for details.
        """
        super().__init__(model)
        self.qoi = qoi_estimator

        if sampler is None:
            sampler = MeanSampler()
        self.sampler = sampler

    @t_batch_mode_transform(expected_q=1)
    def forward(self, X: torch.Tensor) -> torch.Tensor:  # noqa: N803  # pyright: ignore[reportIncompatibleMethodOverride] (necessary due to `t_batch_mode_transform` decorator)
        """Forward method for the QoI acquisition function.

        For each candidate point in x this acquisition function does the following:

        - Fantasizes (via the chosen sampler) possible new observations (y) at this candidate point.
        - Trains a new GP with the additional (x,y_i) pair for each fantasy observation y_i.
        - Evaluates the qoi_estimator on these new GPs, and reports the variance in the resulting estimates.

        This optimization will pick the point that resulted in the lowest mean variance in the QoI estimates.

        Args:
            X: (t_batch, 1, d) input points to evaluate the acquisiton function at.

        Returns:
            The output of the QoI acquisition function that will be optimized with shape (num_points,).

        Todo:
            - This should be updated to use the `FantasizeMixin`. This is the botorch interface indicating
              `model.fantasize(X)` is supported, which does a large chunk of the functionality below. The challenge is
              `model.fantasize(X)` adds an additonal dimension at the start of all posteriors calculated e.g.
              `(num_fantasies, batch_shape, n, m)`. Unclear if our QoI methods can handle/respect the `num_fantasies`
              dimension, of if the different fantasy models can easily be extracted. Revisit at a future date.
            - Consider making the jobs batchable or multiprocessed.
        """
        # shape is: (t_batch, d)
        x = X.squeeze(1)
        x_grad = x.requires_grad

        if x_grad:
            msg = (
                "Gradient tracking is not yet supported, please set `no_grad=True` in optimizer settings."
                "See tutorials/ax_botorch/botrch_minimal_example_custom_acq.ipynb for details"
            )
            raise NotImplementedError(msg)

        # .model.posterior(X) turns gradient back on unless torch.no_grad is used.
        with ExitStack() as stack:
            # sourcery skip: remove-redundant-if
            if not x_grad:
                stack.enter_context(torch.no_grad())

            reject_if_batched_model(
                self.model
            )  # the posterior predictions used in forward don't support batched-models
            posterior = self.model.posterior(x)
            # shape is:  (n_posterior_samples, t_batch, m)
            y = self.sampler(posterior)

            # shape is: (t_batch, m)
            yvar = _get_fantasy_observation_noise(self.model, x)

            # extend the shapes so they all match (n_postior_samples, t_batch, m)
            x = x.expand(y.shape[0], -1, -1)
            yvar = yvar.expand(y.shape[0], -1, -1)

            # shape is: (n_posterior_samples, t_batch)
            lookahead_var = self._batch_lookahead(x, y, yvar)

            # the results of the different y samples need to be aggregated.
            # Use a bespoke aggregator if provided by the sampler, otherwise use the mean
            if hasattr(self.sampler, "mean"):  # noqa: SIM108
                lookahead_var = self.sampler.mean(dim=0)  # type: ignore  # noqa: PGH003
            else:
                lookahead_var = lookahead_var.mean(dim=0)

            # The output from an acquisition function is maximized, so we negate the QoI variances.
            lookahead_var = -lookahead_var

            if x_grad and not lookahead_var.requires_grad:
                msg = (
                    "Losing gradient in QoiLookAhead.forward()\n. Optimisation setting currently require gradient."
                    " This is likely due to gradient not propagated through the qoi_estimator."
                )
                raise RuntimeWarning(msg)

            return lookahead_var

    def _batch_lookahead(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        yvar: torch.Tensor,
    ) -> torch.Tensor:
        """Process a batch of lookahead points.

        Args:
            x (`*b`,n,d): x points to proceess
            y (`*b`,n,m): y points to process
            yvar (`*b`,n,m): yvar points to process

        Returns:
            torch.Tensor (`*b`,n) of lookahead results.
        """
        batch_shape = x.shape[:-2]
        n, d = x.shape[-2:]
        m = y.shape[-1]

        x_flat = x.reshape(-1, d)
        y_flat = y.reshape(-1, m)
        yvar_flat = yvar.reshape(-1, m)

        _results = []
        for x_point, y_point, yvar_point in zip(x_flat, y_flat, yvar_flat, strict=True):
            qoi_var = self.lookahead(x_point, y_point, yvar_point)
            _results.append(qoi_var)

        results = torch.tensor(_results)
        results = results.reshape(*batch_shape, n)
        return results

    def lookahead(self, x_point: torch.Tensor, y_point: torch.Tensor, yvar_point: torch.Tensor | None) -> torch.Tensor:
        """Performs a single lookahead calculation.

        Adds a single additional datapoint to the GP, and determines the QoI with the new GP.

        Args:
            x_point: (d,) The x location of the new point
            y_point: (m,) The y (target) of the new point
            yvar_point: (m,) The y_var  of the new point

        Returns:
            (,) Variance of the QoI with the lookahead GP
        """
        # if homoskedastic noise is used yvar needs to be None due to conditional_upate
        if isinstance(self.model.likelihood, GaussianLikelihood):
            yvar_point = None

        updated_model = conditional_update(
            self.model,
            X=x_point.unsqueeze(0),
            Y=y_point.unsqueeze(0),
            observation_noise=yvar_point.unsqueeze(0) if yvar_point is not None else None,
        )

        #  Calculate the QoI estimates with the fantasized model
        qoi_estimates = self.qoi(updated_model)

        # Calculate the variance of the QoI estimates
        return self.qoi.var(qoi_estimates)


def conditional_update(model: Model, X: torch.Tensor, Y: torch.Tensor, observation_noise: torch.Tensor | None) -> Model:  # noqa: N803
    """A wrapper around `BatchedMultiOutputGPyTorchModel.condition_on_observations` with a number of safety checks.

    This function adds an additional datapoint to the model, preserving the dimension of the original model. Does not
    changing any of the models hyperparameters. This is like training a new `SingleTaskGP` with all the datapoints
    (Hyperparameters are not fit in SingleTaskGP).

    Args:
        model (Model): The model to update.
        X: As per `condition_on_observations`. Shape (`*b`, n', d).

            .. Note::
                `condition_on_observations` expects this to be in the "model" space. It will not be
                transformed by the `input_transform` on the model.

        Y: As per `condition_on_observations`. Shape (`*b`, n', m).

            .. Note::
                `condition_on_observations` expects this to be in the "output/problem" space (not model space).
                It will be transformed by the `output_transform` on the model.

        observation_noise (torch.Tensor | None): Used as the `noise` argument in `condition_on_observations`. Shape
            should match `Y` (`*b`, n', m).

            .. Note::
                `condition_on_observations` expects this to be in the "model" space. It will not be
                transformed by the `output_transform` on the model.

    Return:
        - GP with the same underlying structure, including the new points, and the same original number of dimensions.

    Developer Note:
        There are different ways to create a fantasy model. The following were considered:

        - `BatchedMultiOutputGPyTorchModel.condition_on_observations`: well documented interface producing a GP of
          the same format.
        - `model.get_fantasy_model`: This is a `Gpytorch` implementation. Interface uses different notation, and
          input shape need to be manually adjusted depending on the model.
        - `model.fantasize`: This method would be very convient for our wider purpose, but its posteriors is of shape
          `(num_fantasies, batch_shape, n, m)`. Unclear if our QoI methods can handle/respect the `num_fantasies` dim.

        Revisit this at a later date.
    """
    if not isinstance(model, SingleTaskGP):
        msg = (
            f"Currently only supports models of type SingleTaskGP, received {type(model)}."
            "Currently this has only been tested for SingleTaskGP (with Homoskedastic and Fixed noise)."
            " It can likely be expanded to broader classes of model, but it is important to understand the noise"
            "expected by the model, as this can silently create incorrect GPs if incorrect. See TODO(link unit test)"
            " for details. As such, currently only GPs that have been explicitly tested are accepted."
        )
        raise NotImplementedError(msg)

    if isinstance(model.likelihood, FixedNoiseGaussianLikelihood):
        if observation_noise is None:
            msg = "Conditional update of SingleTaskGP with FixedNoise requires observation_noise to be provided."
            raise ValueError(msg)
    elif isinstance(model.likelihood, GaussianLikelihood):
        if observation_noise is not None:
            msg = (
                "Conditional update of the observation_noise is not supported for SingleTaskGPs with Homoskedastic "
                "Noise. This combination leads to inconsistent updated of the GP."
            )
            raise ValueError(msg)
    else:
        msg = (
            "Expected the model.likelihood to be either GaussianLikelihood or FixedNoiseGaussianLikelihood."
            " Recieved {model.likelihood}"
        )
        raise TypeError(msg)

    if hasattr(model, "input_transform") or hasattr(model, "outcome_transform"):
        msg = (
            "Caution: the model passed has input or outcome transforms. Need to be extremely careful that all inputs"
            "the in the correct space ('model space' vs 'problem/outcome space'). Please see docstring for more"
            " details."
        )
        warnings.warn(msg, stacklevel=5)

    if model.batch_shape != X.shape[:-2] or model.batch_shape != Y.shape[:-2]:
        msg = (
            f"Model.batch_shape is {model.batch_shape}, but X batch_shape is {X.shape[:-2]} and Y batch_shape is:"
            f" {Y.shape[:-2]}. The resultant model will not have the same batch shape as the original."
        )
        raise ValueError(msg)

    # Botorch required the model be have been run (so caches have been populated) before conditioning can be done.
    # This is the method Botrch used to do this. We run the model on a single input if required
    if model.prediction_strategy is None:
        _ = model.posterior(X=X)

    new_model = model.condition_on_observations(
        # MUST BE IN MODEL SPACE - will not be automatically transformed
        # shape require (b,n',d)
        X=X,
        # Must be in outcome space - will be automatically transformed
        # Shape require (b,n',m)
        Y=Y,
        # Must be in model space - - will not be automatically transformed
        # shape must be the same as Y
        noise=observation_noise,
    )

    return new_model


# TODO(sw 2024-11-27): there are cases where users might want to chose between closest_observational_noise and
# average_observational_noise. This should be a parameter of QoILookAhead.
def closest_observational_noise(
    new_points: torch.Tensor, train_x: torch.Tensor, train_yvar: torch.Tensor
) -> torch.Tensor:
    """Find the closest point in a training dataset, and collect its observational noise.

    Args:
        new_points: (n, d) The points to produce observational_noise. Features should be normalise to [0,1] square.
        train_x: (n',d) The points to compare similarity to. Features should be normalise to [0,1] square.
        train_yvar: (n',m) The observational variance associated with each point.

    Return:
        (n,m) Tensor with the variance for each of the new_points.


    Details:
    This function is useful for Non-Batched SingleTaskGPs because they will always have arguments of these dimension.

    Warning:
        This function is not smooth, meaning optimizers that use gradient (1st or 2nd order derivatives) such as
        L-BFGS-B will not work. The trade off is it is more robust to the effect of patterns in yvar than
        `average_observational_noise` . See Issue #213 for details.

    """
    # NOTE: Training data might be less than [0,1] because it can be standardised with preset bounds (e.g env bounds)
    # new_points could potentailly be outside of bounds if bounds were determined by the data.
    if train_x.min() < 0 or train_x.max() > 1:
        # TODO(sw 2024-11-15): We could use normalise to standardise the model here ourselves.
        msg = (
            "The models train_x data is not standardised to [0,1]. Non-standarised features will biases large"
            " features in the similarity measure used in this function."
            f" Found min={train_x.min()}, max={train_x.max()}."
        )
        warnings.warn(msg, stacklevel=8)

    # measure the distance from new_points to train_points to find the one that is most similar.
    # (new_point, distance to points)
    distances = torch.cdist(new_points, train_x, p=2)
    closest_point_idx = distances.argmin(dim=-1)
    return train_yvar[closest_point_idx]


def average_observational_noise(
    new_points: torch.Tensor,
    train_x: torch.Tensor | None,  # noqa: ARG001
    train_yvar: torch.Tensor,
) -> torch.Tensor:
    """Return the average observational noise.

    Args:
        new_points: (n, d) The points to produce observational_noise.
        train_x: (n',d). This is not used, but is kept to keep a consistent function signature
        train_yvar: (n',m) The observational variance associated with each point.

    Return:
        (n,m) Tensor with the variance for each of the new_points.

    Details:
    This function is useful for Non-Batched SingleTaskGPs because they will always have arguments of these dimension.

    Warning:
        Certain pattern of homoskedasticity cause this method to perform poorly, causing the acquisition function to
        recommend suboptimal points (as compared to `closest_observational_noise`). The trade off is derivative based
        optimisation techniques can be used. See Issue #213 for details.
    """
    return train_yvar.mean(dim=0).expand(new_points.shape[0], -1)


def _get_fantasy_observation_noise(model: SingleTaskGP, x: torch.Tensor) -> torch.Tensor:
    """Estimates the observation noise of a new point by finding the closest point in a model.

    Distance is assumed a good measure of point similarity.

    Args:
        model: The model from which to use the training data.
        x: (n,d) The points to find fantasy noise for. Batching is not supported. This should be in model space.

    Return:
        Tensor (n,m) of observation noise (variance) for each point in x.

    Todo TODO:
        - Review the implication noise picking choice on optimisation. See Issue #213 for details.
    """
    reject_if_batched_model(model)

    # Gaurenteeed to be (n,d)
    model_train_x = get_training_data_singletaskgp(model)
    # # Gaurenteeed to be (n,m) due to rejecting batched_models
    model_yvar = get_target_noise_singletaskgp(model)

    # if model uses homoskedastic noise output will need expand output from (1,m) to (n,m)
    model_yvar = model_yvar.expand(model_train_x.shape[0], -1)

    x_obs_noise = average_observational_noise(x, model_train_x, model_yvar)

    return x_obs_noise


def reject_if_batched_model(model: SingleTaskGP) -> None:
    """Helper function to reject batched model in code where they are not yet supported.

    Args:
        model: The model to check

    Return:
        Raise not yet implements in model is batched. Otherwise None.

    Details:
        botorch models can have batched training data, and or batched.

        - gp batch prediction (non-batched model):

          - train_x = (n,d) # This is a single GP
          - train_y = (n,m)
          - predicting_x = (b,n',d)
          - result will be: (b, n',m).

            - There are b seperate joint distribution (each with n points, a t targets)

        - batched gps model:

          - train_x = (b_gp,n,d)
          - train_y = (b_gp,n,m)

            - b_gp seperate GPs, where each GP gets all its own hyperparams etc trained on (n,d) point.

          - prediciton_x = (n',d)
          - result will be: (b_gp, n',m).

            - Each of the seperate b_gp gps makes its own estimate of the joint distribution.

        More details: `BoTorch batching <https://botorch.org/docs/batching>`_
    """
    if len(model.batch_shape) != 0:
        msg = (
            "batch models are not currently supported. Model has training data (*b,n,d), on (n,d) is supported."
            "See https://botorch.org/docs/batching for more details on batching"
        )
        raise NotImplementedError(msg)


@acqf_input_constructor(QoILookAhead)
def construct_inputs_qoi_look_ahead(  # noqa: D417
    model: SingleTaskGP, qoi_estimator: QoIEstimator, sampler: PosteriorSampler, **_: dict[str, Any]
) -> dict[str, Any]:
    """This is how default arguments for acquisition functions are handled in Ax/Botorch.

    Context
        When Ax.BOTORCH gets instantiated, construction arguments for the acquisition function can be provided.
        These are passed through Ax as a set of Kwargs

    Args:
        This function takes a subset of the acquisition functions `__init__()` args and can add defaults.

    Returns:
        Args for the Botorch acquisition function `__init__()` (output).

    Note:
        This functionality allows Ax to pass generic arguments without needing to know which acquisition function they
        will be passed to. Interestingly, this functionality is provided by the BoTorch package, even though it seems
        like it should be the responsibility of Ax. This issue is discussed in detail here: `GitHub discussion <https://github.com/pytorch/botorch/discussions/784>`_.
    """
    return {
        "model": model,
        "qoi_estimator": qoi_estimator,
        "sampler": sampler,
    }


@optimizer_argparse.register(QoILookAhead)
def _argparse_qoi_look_ahead(
    # NOTE: this is a bit of a non-standard implementation because we need to update params in a nested dict. Would
    #  prefer to set the key values directly here
    acqf: QoILookAhead,
    # needs to accept the variety of args it is handed, and then pick the relevant ones
    **kwargs: Any,  # noqa: ANN401
) -> dict[str, Any]:
    """Controls the default optimization parameters used.

    The following provides an overview of how passing optimization argument to `ax.TorchModelBridge.gen()`.
    Anything that is not passed here will used the default value defined by this function.

    ```python
        TorchModelBridge.gen(
            ...                                 # Other parameters there that are not relevant
            model_gen_options: {
                optimise_kwargs: {              # content of this key later used like `botorch.optim.optimize_acqf(**optimise_kwargs)`  # noqa: E501
                    options: {                  # This is a parameter of `botorch.optim.optimize_acqf`
                        with_grad: False        # Example of some optimization params that can be passed
                        ...
                        method: "L-BFSG"
                    }
                }
            }
    ```

    Rough flow of this object through the stack:
        - `ax.TorchModelBridge.gen(..., model_gen_options: dict)`
        [internally end up calling `TorchModelBridge.model` giving the below object]
        - `ax.BotorchModel.gen(..., torch_opt_config: TorchOptimConfig)`
            - Same way TorchModelBridge convert x data from ax type to tensors, the config type is converted.
        [Internally that leads to the following being called]
        - `ax.Acquisition.optimise(..., opt_options)`
            - where `opt_options = torch_opt_config.model_gen_options.optimize_kwargs`
        [Internally this leads to the following being called]
        - `botorch.optim.optimize_acqf(..., opt_options_with_defaults)`
            - `opt_options_with_defaults` is `opt_options` with defaults added.

    This function helps add the defaults in the final step

    Note:
        - This function is stored in a registry in Ax. This registery is searched (by acquisition class name)
          when the acqf func is being optimized.
        - Any variable passed from `.gen` must override the defaults specified here
        - Existing default implementations can be found here: ax.models.torch.botorch_modular.optimizer_argparse

    Args:
        acqf: Acqusiton that will be passed.
        kwargs:
            - everything passed in value of `optimizer_kwargs` in `model.gen(model_gen_options = {optimizer_kwargs ={}})` can be found with
            `kwargs[optimizer_options]`

    Returns:
        dict of args unpacked to `botorch.optim.optimize_acqf`.
            - All args are set with this, excluding the following:
                - `acq_function`,`bounds`,`q`,`inequality_constraints`,`fixed_features`,`post_processing_func`
    """  # noqa: E501
    # Start by using the default arg constructure, then adding in any kwargs that were passed in.

    # NOTE: this is an internal method, using it is an anti pattern.
    # - Ax just stores all the arge parsers in the same file as _argparse_base so they can use it directly.
    # - We can't put this function in that file, even though it belongs there.
    # Definition of _argparse_base explains the shape returned
    args = _argparse_base(acqf, **kwargs)

    # Only update with these defaults if the variable were not passin in kwargs
    optimizer_options = kwargs["optimizer_options"]
    options = optimizer_options.get("options", {})

    if "raw_samples" not in optimizer_options:
        args["raw_samples"] = 100
    if "with_grad" not in options:
        args["options"]["with_grad"] = False

    if "method" not in options:
        args["options"]["method"] = "Nelder-Mead"
        # These options are specific to using Nelder-Mead
        if "maxfev" not in options:
            args["options"]["maxfev"] = "5"
        if "retry_on_optimization_warning" not in optimizer_options:
            args["retry_on_optimization_warning"] = False

    return args



"""Module for the MeanPosteriorSampler class."""

import torch
from botorch.posteriors import GPyTorchPosterior

from axtreme.sampling.base import PosteriorSampler


class MeanSampler(PosteriorSampler):
    """Class for sampling a posterior by returning the mean of the posterior.

    This is a simple sampler that just returns the mean of the posterior.
    """

    def __call__(self, posterior: GPyTorchPosterior) -> torch.Tensor:
        """Return the mean of the posterior."""
        # Unsqueezed first dimension so it conforms to the required dimension.
        # TODO(sw): does this conform for multi and single variate
        return posterior.mean.unsqueeze(0)



from typing import Any

import numpy as np
import pytest
from scipy import stats
from scipy.stats._distn_infrastructure import rv_continuous

from axtreme.utils import distibution_helpers


@pytest.mark.parametrize(
    "dist, expected",
    [
        (stats.gumbel_r, ["loc", "scale"]),
        (stats.weibull_max, ["c", "loc", "scale"]),
        (stats.beta, ["a", "b", "loc", "scale"]),
    ],
)
def test_distribution_parameter_names_from_scipy(dist: rv_continuous, expected: list[str]):
    result = distibution_helpers.distribution_parameter_names_from_scipy(dist)
    assert result == expected


# TODO(sw): How do we chance that the uncertainty (cov) we are getting is reasonable?
@pytest.mark.parametrize(
    "dist, params",
    [
        pytest.param(stats.gumbel_r, {"loc": 4, "scale": 1}, id="gumbel_r"),
        pytest.param(
            stats.weibull_max,
            {"c": 1, "loc": 4, "scale": 1},
            marks=pytest.mark.xfail(reason="Currently only support gumbel"),
            id="weibull_max",
        ),
    ],
)
def test_fit_dist_with_uncertainty(dist: rv_continuous, params: dict[str, Any]):
    sample = dist.rvs(**params, size=10_000, random_state=7)
    means, _ = distibution_helpers.fit_dist_with_uncertainty(sample, dist)
    np.testing.assert_allclose(means, list(params.values()), atol=1e-2)
